from myraptor.helpers import *
#from helpers import *

def raptor(tb,start_stop_name,target_stop_name,start_time,max_transfers):
    """
    Function representing the myraptor algorithm
    Parameters: tb (Timetable object)
        start_stop_name (str)
        target_stop_name (str)
        start_time (datetime object)
        max_transfers (int)
    Returns :Fastest path from the start_stop to the target_stop (tuple (path,
    start_stop, target_stop))

    """
    #Creation of stop objects for the provided stop arguments
    start_stop = tb.get_stop(start_stop_name)
    target_stop = tb.get_stop(target_stop_name)

    #RAPTOR ALGORITHM
    # 1) Initialization : Update of start_station attribute best arrival time
    #                     Building Queue of updated stops

    start_stop.best_arr = start_time
    stop_queue = [start_stop]
    current_time = start_time
    # 2) First loop: represents rounds in the algorithm, while each round
    #                computes the fastest way of getting to every stop with
    #                at most k-1 transfers
    for transfer in range (max_transfers+1):
        stop_queue_previous = stop_queue
        while stop_queue_previous:
    # 3) First part of round: sets the current time to the best arrival time
    #    from the previous round
            stop_queue = []
            current_stop = stop_queue_previous.pop()
            current_time = current_stop.best_arr
            route_queue = [tb.get_route(route) for route in current_stop.routes]
    # 4) Second part of the round: processing of the routes
            while route_queue:
                route = route_queue.pop()

                #selection of data departure data concerning only current stop
                dep_current_stop = route.create_departure_list(current_stop)
                #selection of best departure time from current stop
                if dep_current_stop == []:
                    continue
                best_time =binary_search(dep_current_stop,current_time)
                dep = best_time[0]
                dep_index = best_time[1]
                if dep == None:
                    continue
                # selection of trip with best departure time
                current_trip = route.trip_ids[dep_index]
                # decision if the arrival time to the stop was improved
                for stop_name in route.get_rest_stops(current_stop):
                    stop = tb.get_stop(stop_name)
                    arr = current_trip.get_arr(stop)
                    if arr < stop.best_arr:
                        stop.improve_stop( arr, dep, current_stop, route)
                        if stop in stop_queue:
                            continue
                        stop_queue.append(stop)


    if target_stop.marked == False:
        print ("It is not possible to find the connection, try more"
               " transfers!")
        return None
    return (path_to_str(target_stop.get_path(),start_stop,target_stop))


import pkgutil
from copy import deepcopy
from myraptor.helpers import *
#from helpers import *
from myraptor.classes import *
#from classes import *
from myraptor.raptor import raptor
#from raptor import raptor
from myraptor.create_tb import *


source_data = 'data/data.csv'
source_data_stops = 'data/data_stops.csv'

def main():
    timetable = create_tb(open_data(source_data),
                         open_data_stops(source_data_stops))
    while True:
        while True:
            tb = deepcopy(timetable)

            start_stop_name = load_start_stop(tb)
            target_stop_name = load_target_stop(tb)
            start_time_datetime = load_start_time()
            start_time = time_to_seconds(start_time_datetime)
            transfers = load_transfers()
            raptor(tb, start_stop_name, target_stop_name, start_time,
                   transfers)

            continue_or_no = load_yes_no()
            if continue_or_no == "y":
                continue
            elif continue_or_no == "n":
                print("Thank you for using myraptor!")
                break



if __name__ == '__main__':
    main()





import datetime as dt
from datetime import datetime


def binary_search(dep_times_list, start_time):
    """
    Function representing the binary search algorithm, from a list of ordered
    values (departure times) chooses value equal or closest (higher) to given
    parameter
    Parameters
    ----------
    dep_times_list: list of ordered datetime objects
    start_time: datetime object

    Returns: tuple of selected departure time and index of the time
    -------

    """
    #if start time earlier than all departure times, the first departure time
    #is chosen
    if (start_time <= dep_times_list[0]):
        return (dep_times_list[0], 0)
    #if start time is after all departure times, no connection is available
    #within this route, the best time is set to None
    if start_time > dep_times_list[len(dep_times_list) - 1]:
        return None,None
    left_index = 0
    right_index = len(dep_times_list) - 1

    while left_index <=right_index:
        if left_index == right_index:
            return (dep_times_list[left_index], left_index)
        middle_index = (left_index + right_index) // 2
        middle_time = dep_times_list[middle_index]
        if start_time > middle_time:
           left_index = middle_index + 1
        elif start_time < middle_time:
           right_index = middle_index
        else:
            return (dep_times_list[middle_index], middle_index)
    return (None, None)

def path_to_str(path, start_stop, target_stop):
    """ Prints the path from the start_stop to the target_stop in
     readable format"""
    for station in path:
        if station[0] == start_stop.stop_name:
            stop_name = station[0]
            departure_time = time_to_hhmmss(station[2])
            line = station[3]
            path_str = (
                f"stop: {stop_name},  "
                f"departure time: {departure_time},  "
                f"take line: {line}  "
            )
        elif station[0] == target_stop.stop_name:
            stop_name = station[0]
            arrival_time = time_to_hhmmss(station[1])
            path_str = (
                f"stop: {stop_name},  "
                f"arrival time: {arrival_time}  "
            )
        else:
            stop_name = station[0]
            arrival_time = time_to_hhmmss(station[1])
            departure_time = time_to_hhmmss(station[2])
            line = station[3]
            path_str = (
                f"stop: {stop_name},  "
                f"arrival time: {arrival_time},  "
                f"departure time: {departure_time},  "
                f"change to line: {line}  "
            )
        print(path_str)


def time_to_hhmmss(time):
    """Transfers time in form of total seconds to timedelta object in form
    of HH:MM:SS"""
    return str(dt.timedelta(seconds=time))

def time_to_str(time):
    """Receives timedelta object and transforms it to string
     in form HH:MM:SS"""
    return time.strftime("%H:%M:%S")



def load_start_stop (tb):
    """Asks the user to type start_stop. start stop must be in the list
        tb.stop_names. If not, the list of tb.stop_names is printed """
    while True:
        start_stop_name = input("Please enter start station: ")
        if start_stop_name in tb.stop_names:
            return start_stop_name
        else:
            print ("Not valid name of the station, please start again")
            print (sorted(tb.stop_names))


def load_target_stop (tb):
    """Asks the user to type target_stop. start stop must be in the list
        tb.stop_names. If not, the list of tb.stop_names is printed """
    while True:
        target_stop_name = input("Please enter target station: ")
        if target_stop_name in tb.stop_names:
            return target_stop_name
        else:
            print ("Not valid name of the station, please start again")
            print (sorted(tb.stop_names))



def load_start_time ():
    """Asks the user to type start_time in form HH:MM:SS """
    while True:
        start_time = input(
            "Please enter start time in form HH:MM:SS   :   ")
        try:
            time = datetime.strptime(start_time, '%H:%M:%S')
            return time
        except UnicodeDecodeError:
            start_time.encode("latin-1", "ignore")
            print (start_time)
            time = datetime.strptime(start_time, '%H:%M:%S')
        except:
            print ("Not valid format of time")


def time_to_seconds(time):
    """transfers datetime object HH:MM:SS and transfers it to seconds (float)"""
    time_zero = datetime.strptime('00:00:00', '%H:%M:%S')
    time_total = time - time_zero
    time_total = time_total.total_seconds()
    return time_total

def load_transfers ():
    """Asks the user to type maximum number of transfers allowed"""
    #TODO: check if the number is right
    while True:
        transfers = input("Please enter maximum allowed transfers in form of"
                          " number in range (0,10): ")
        try:
            return int(transfers)
        except UnicodeDecodeError:
            print("unicode")
            transfers.encode("latin-1", "ignore")
            return(int(transfers))
        except ValueError:
            print ("This is not a number")

def load_yes_no ():
    """Asks the user if he wants to find another connection, answer must
    be in form of str 'y' or 'n'"""
    while True:
        answer =input("Do you want to find another connection? y/n: ")
        if answer == "y":
            return answer
        elif answer[0] == "n":
            return answer
        else:
            print ("Please write y or n!")






from io import BytesIO
import pandas as pd
import numpy as np
import datetime
from datetime import datetime
import sys
import pkgutil
from myraptor.classes import *
#from classes import *

source_data = 'data/data.csv'
source_data_stops = 'data/data_stops.csv'


def open_data_stops(source_data_stops):
    """Opens file containing dataset  concerning stops. Dataset was prepared
     in jupyter notebook data_raptor"""
    print ("Loading data, please wait....")
    data_open = pkgutil.get_data(__name__,source_data_stops)
    data_stops = pd.read_csv(BytesIO(data_open),
                             converters={'stop_routes_all': eval})
    return data_stops

def open_data(source_data):
    """Opens file containing dataset  concerning routes and trips. Dataset was
     prepared in jupyter notebook data_raptor"""
    data_open = pkgutil.get_data(__name__, source_data)
    data = pd.read_csv(BytesIO(data_open),
                       converters={'trip_arrival_times_seconds': eval,
                                   'trip_departure_times_seconds': eval,
                                   'trip_stop_names': eval})
    return data

def create_tb(data,data_stops):
    """Creates Timetable objects representing Prague public
    transport network (MHD). It contains Trip objects, Routes objects and
    Stop objects.
    Each row in prepaired dataset "data" contains data about one trip.
    Each row in prepaired dataset "data_stops" contains data about one stop."""
    print ("Creating timetable, please wait....")
    all_raptor_routes = data.route_raptor_id.unique()
    time1 = datetime.strptime('23:59:59','%H:%M:%S')
    tb = Timetable()
    #Creation of Trip objects (trip_stop_times list) and
    # Route objects (route_trips list)
    for routing in all_raptor_routes:
        #choose only data with the raptor_id
        data_route = data.loc[data["route_raptor_id"]==str(routing)]
        if routing == '':
            continue
        new_route = tb.create_route(routing)
        stops = data_route.iloc[0]["trip_stop_names"]
        name = data_route.iloc[0]["route_short_name"]
        new_route.stop_names = stops
        new_route.name = name
        for row in data_route.itertuples():
            if len(row.trip_departure_times_seconds) != len(
                    new_route.stop_names):
                break
            new_trip = tb.create_trip (row.trip_id,routing)
            new_trip.stop_times_arr = row.trip_arrival_times_seconds
            new_trip.stop_times_dep = row.trip_departure_times_seconds
            new_trip.number_stops +=1
            new_route.number_trips +=1
    #Creation of Stop objects (list stop_routes)
    for row in data_stops.itertuples():
        new_stop = tb.create_stop(row.stop_name)
        routes = row.stop_routes_all
        list_routes = []
        unique_routes = set(routes)
        for route in unique_routes:
            list_routes.append(route)
        new_stop.routes = list_routes
    return (tb)




import datetime
from datetime import datetime
import sys



class Timetable:
    """
    Class representing timetable, with routes, trips and stops.
    Attributes:
        routes: dictionary (contains all routes
                            key: raptor_id,
                            value: Route object)
        stops: dictionary (contains all stops,
                            key: stop_name,
                            value: Stop object)
        stop_names : list, contains all stops names

    """

    def __init__(self):
        self.routes = {}
        self.stops = {}
        self.stop_names = []

    def create_route(self, raptor_id):
        """
        Creates Route object
        Parameter: raptor_id (str)
        Returns: Route object
        """
        new_route = Route(raptor_id)
        self.routes[raptor_id] = new_route
        return new_route

    def get_route(self, raptor_id):
        """
        Recieves myraptor id (str), returns corresponding Route object
        Parameter: raptor_id (str)
        Returns: Route object
        """
        return self.routes[raptor_id]

    def create_trip(self, trip_id, raptor_id):
        """
        Creates Trip object
        Parameters: raptor_id (str)
                    trip_id (str)
        Returns: Trip object
        """
        route = self.get_route(raptor_id)
        new_trip = Trip(trip_id, route)
        route.trip_ids.append(new_trip)
        return new_trip

    def create_stop(self, stop_name):
        """
        Creates Stop object
        Parameter: stop_name (str)
        Returns: Stop object
        """
        new_stop = Stop(stop_name)
        self.stops[stop_name] = new_stop
        self.stop_names.append(stop_name)
        return new_stop


    def get_stop(self, stop_name):
        """
        Recieves stop name (str), returns corresponding Stop object
        Parameter: stop_name (str)
        Returns: Stop object
        """
        return self.stops[stop_name]


class Trip:
    """Class representing Trips. A trip is a sequence of two or more stops that
     occur during a specific time period.
    Attributes:
        trip_id
        number_of_stops
        route: route on which the trip operates
        stop_times_arr,stop_times_dep list, ordered sequence of arrival
         and departure times at all stops the trip contains
        (arrival_stop1,arrival_stop2,arrival_stop3,....)
    """
    def __init__(self, trip_id, route):
        self.trip_id = trip_id
        self.number_stops = 0
        self.route = route
        self.stop_times_arr = []
        self.stop_times_dep = []

    def get_stop_index(self, stop):
        """Returns index of given stop in the list route.stop_names
        operates
        Parameter: Stop object
        Returns: index (int)"""
        index = self.route.get_stop_index(stop)
        return index

    def get_arr_index(self, stop):
        """
        Returns index of arrival time to given stop in the list stop_times
        Parameter: Stop object
        Returns: index (int)
        """
        index = self.get_stop_index(stop)
        return index

    def get_arr(self, stop):
        """
        Returns  arrival time to given stop
        Parameter: Stop object
        Returns: arrival_time (datetime object)
        """
        return self.stop_times_arr[self.get_arr_index(stop)]

    def get_dep_index(self, stop):
        """
        Returns index of departure time from given stop in the list stop_times
        Parameter: Stop object
        Returns: index (int)
        """
        index = self.get_stop_index(stop)
        return index

    def get_dep(self, stop):
        """
        Returns  departure time from given stop
        Parameter: Stop object
        Returns: departure_time (datetime object)
        """
        dep = self.stop_times_dep[self.get_dep_index(stop)]
        return dep


class Route:
    """
    Class representing routes. A route is a group of trips that contains
    exactly same stops in same order.
    Attributes:
        raptor_id
        name : name of the line, eg. metro A, tram 22 etc.
        number_stops
        number_trips
        stop_names: list, ordered sequence of stop names of the route
        trip_ids: list, ordered list of trip ids operating the route
    """
    def __init__(self, raptor_id):
        self.raptor_id = raptor_id
        self.name = None
        self.number_stops = 0
        self.number_trips = 0
        self.stop_names = None
        self.trip_ids = []


    def get_stop_index(self, stop):
        """Returns index of given stop in the list stop_names
                operates
                Parameter: Stop object
                Returns: index (int)"""
        index = self.stop_names.index(stop.stop_name)
        return index

    def get_stop_sequence(self, stop):
        """Returns sequence of given stop in the list stop_names
                operates
            Parameter: Stop object
            Returns: sequence (int)"""
        sequence = self.get_stop_index(stop.stop_name) + 1
        return sequence

    def create_departure_list(self, stop):
        """Returns list of upward ordered departures from given stop of all
         trips belonging to the route
         Parameters: Stop object
         Returns departures (list)"""
        departures = []
        for trip in self.trip_ids:
            dep = trip.get_dep(stop)
            departures.append(dep)
        return departures



    def get_rest_stops(self, stop):
        """Returns ordered list of stop_names which follow after the given
        stop
        Parameter: Stop object
        Returns resto_stops (list)"""
        stop_index = self.stop_names.index(stop.stop_name)
        rest_stops = self.stop_names[(stop_index + 1):]
        return rest_stops

class Stop:
    """Class representing Stops
        Attributes:
            stop_name (str)
            routes (list, containes all routes serving the stop)
            best_arr (datetime object, set to max, changed if improved)
            dep_time (datetime object)
            marked (bool, marked = True if the stations best_arr was improved)
            self.previous_stop (Stop object)
            self.previous_line (str)
            self.previous_dep (datetime object)
            """
    def __init__(self, stop_name):
        self.stop_name = stop_name
        self.routes = []
        self.best_arr = sys.maxsize
        self.dep_time = None
        self.marked = False
        self.previous_stop = None
        self.previous_line = None
        self.previous_dep = None

    def get_path(self):
        """Returns path, how to get to the stop from the start_stop in form
        of list of tuples. Each tuple represents one stop from the path.The
        list contains exactly one start_stop tuple,any number of middle_stop
        tupples and exactly one target_stop tupple.
        Start_stop tuple contains start_stop_name, departure_time, line,
        middle_stops tuple contains
                        stop_name, arrival_time, departure_time, line
        target_stop tuple contains target_stop_name, arrival time."""
        path = [(self.stop_name,self.best_arr,None,None)]
        stop=self
        while stop.previous_stop:
            #(stop_name, arrival time, departure_time, line)
            path.append((stop.previous_stop.stop_name,
                         stop.previous_stop.best_arr,
                         stop.previous_dep,
                         stop.previous_line))
            stop = stop.previous_stop
        path.reverse()
        return path

    def improve_stop(self,arr,dep,current_stop,route):
        """ Executes changes to the Stop object attributes: improves arrival
        and departure times and sets previous stop and line"""
        self.best_arr = arr
        self.dep_time = dep
        self.previous_stop = current_stop
        self.previous_dep = dep
        self.previous_line = route.name
        self.marked = True



from setuptools import setup

def get_long_description():
    with open('README.md') as f:
        long_description = f.read()

REQUIREMENTS = ['requests']

setup(name='myraptor',
      version='3.0.0',
      description='Prague public transport connection finder using RAPTOR algorithm',
      long_description=get_long_description(),
      url='https://github.com/hanjak/connection_finder_prague_mhd',
      author='Hana Kosova',
      author_email='hanja.kosova@gmail.com',
      license='MIT',
      packages=['myraptor'],
      install_requires=['pandas==1.3.2','numpy==1.21.2'],
      keywords='raptor connection finder',
      include_package_data=True,
      package_data={'myraptor':['data/data.csv','data/data_stops.csv']},
      )


from copy import deepcopy
from helpers import *
from classes import *
from create_tb import *
#from raptor import raptor

from raptor import raptor

source_data = 'data/data.csv'
source_data_stops = 'data/data_stops.csv'

def main():
    timetable = create_tb(open_data(source_data),open_data_stops(source_data_stops))
    #print (tb)

    while True:
        tb = deepcopy(timetable)

        start_stop_name = load_start_stop(tb)
        target_stop_name = load_target_stop(tb)
        start_time_datetime = load_start_time()
        start_time = time_to_seconds(start_time_datetime)
        transfers = load_transfers()
        raptor(tb,start_stop_name,target_stop_name,start_time,transfers)

        continue_or_no = load_yes_no()
        if continue_or_no == "y":
            continue
        elif continue_or_no == "n":
            print ("Thank you for using myraptor!")
            break








if __name__ == '__main__':
    main()
